import asyncio
import random
import traceback
from collections import namedtuple
from importlib import import_module

import rx
import rx.operators as ops
from rx.scheduler.eventloop import AsyncIOScheduler

from cyclotron.backpressure import pid
from cyclotron.debug import trace_observable
import cyclotron_aiokafka as kafka

from .util import import_function


def initialize_topics(config_topics):
    ''' initializes the topic objects from a configuration

    If no encoding os provided, then the string encoder is used by default.
    '''
    Topic = namedtuple('Topic', ['encode', 'decode', 'map_partition', 'start_from'])

    topics = {}
    for topic in config_topics:
        if "encoder" in topic:
            module = import_module(topic['encoder'])
        else:
            module = import_module("makinage.encoding.string")
        encoder = getattr(module, "encoder")
        encode, decode = encoder()

        if "partition_selector" in topic:
            map_partition = import_function(topic['partition_selector'])
        else:
            def r(i): return int(random.random() * 1000)
            map_partition = r

        if "start_from" in topic:
            start_from = topic['start_from']
        else:
            start_from = 'end'

        topics[topic['name']] = Topic(
            encode=encode,
            decode=decode,
            map_partition=map_partition,
            start_from=start_from,
        )

    return topics


def initialize_regulators(config, kafka_feedback):
    regulators = {}
    for regulator in config:
        control = kafka_feedback.pipe(
            trace_observable("regulator feedback"),
            ops.filter(lambda i: i[0] == regulator['feedback']),
            ops.map(lambda i: i[1] / 1000),

            pid(rx.concat(rx.just(1.0), rx.never()),
                -0.001, -0.001, 0.0),

            #ops.map(lambda i: 1/i if i != 0 else 1.0),
            ops.map(lambda i: max(min(i, 0.01), 0.0)),
            trace_observable("regulator"),
        )

        regulators[regulator['control']] = control

    return regulators


def create_source_observable(kafka_source, source):
    return kafka_source.pipe(
        trace_observable(prefix="kafka source {}".format(source)),
        ops.filter(lambda i: i.topic == source),  # ConsumerRecords
        trace_observable(prefix="kafka consumer records {}".format(source)),
        ops.flat_map(lambda i: i.records),  # ConsumerRecord
    )

def create_operators(config, config_source, kafka_source, kafka_feedback):
    ''' creates the operators declared in config

    Args:
        config: a dict containing the configuration file. todo: observable.
        kafka_source: The kafka response observable
        kafka_feedback: The kafka backpressure process feedback
    Returns:
        An observable containing tuples of (topic, observable).
    '''
    try:
        topics = initialize_topics(config['topics'])
        if 'regulators' in config:
            regulators = initialize_regulators(
                config['regulators'],
                kafka_feedback)
        else:
            regulators = {}
        producers = []
        consumers = []
        for k, operator in config['operators'].items():
            factory = import_function(operator['factory'])
            sources = []
            if 'sources' in operator:
                for source in operator['sources']:
                    print('create source {}'.format(source))
                    consumers.append(kafka.ConsumerTopic(
                        topic=source,
                        decode=topics[source].decode,
                        control=regulators[source] if source in regulators else None,
                        start_from=topics[source].start_from,
                    ))

                    sources.append(create_source_observable(kafka_source, source))

            print(sources)
            sinks = factory(config_source, *sources)
            print("sinks: {}".format(sinks))
            if 'sinks' in operator:
                for index, sink in enumerate(operator['sinks']):
                    print('create sink {} at {}'.format(sink, index))
                    producers.append(kafka.ProducerTopic(
                        topic=sink,
                        records=sinks[index],
                        map_key=lambda i: None,
                        encode=topics[sink].encode,
                        map_partition=topics[sink].map_partition
                    ))

        kafka_sink = []
        if len(consumers) > 0:
            kafka_sink.append(kafka.Consumer(
                server=config['kafka']['endpoint'],
                group=config['application']['name'],
                topics=rx.from_(set(consumers)),
            ))

        if len(producers) > 0:
            kafka_sink.append(kafka.Producer(
                server=config['kafka']['endpoint'],
                topics=rx.from_(producers)
            ))

        kafka_sink = rx.from_(kafka_sink) if len(kafka_sink) > 0 else rx.never()
        return kafka_sink
    except Exception as e:
        print("Error while creating operators: {}, {}".format(
              e, traceback.format_exc()))



import pickle


def encoder():
    ''' Pickle encoder

    encodes any object using the pickle format
    '''
    def _encode(i):
        return pickle.dumps(i)

    def _decode(i):
        return pickle.loads(i)

    return _encode, _decode



import rx

import cyclotron_std.argparse as argparse
import makinage.config as config


def test_parse_arguments():
    source = rx.from_(['makinage', '--config', 'test.yaml'])
    actual_result = []
    actual_error = []
    config.parse_arguments(source).subscribe(
        on_next=actual_result.append,
        on_error=actual_error.append,
    )

    assert actual_result == [argparse.Argument(key='config', value='test.yaml')]



import yaml

import rx
import rx.operators as ops

import cyclotron_std.sys.argv as argv
import cyclotron_std.argparse as argparse
import cyclotron_std.io.file as file
import cyclotron_consul.kv as kv


def parse_config():
    def _parse_config(config_data):
        ''' takes a stream with the content of the configuration file as input
        and returns a (hot) stream of arguments .
        '''
        return config_data.pipe(
            ops.map(lambda i: yaml.load(i, Loader=yaml.FullLoader)),
            ops.replay(buffer_size=1),
            ops.ref_count(),
        )

    return _parse_config


def parse_arguments(args, prog=None):
    parser = argparse.ArgumentParser(prog=prog)
    parser.add_argument(
        '--config', required=True,
        help="Path of the configuration file")

    return args.pipe(
        ops.skip(1),
        argparse.parse(parser),
    )


def read_config_from_file(filename, file_response, scheduler=None):
    read_request, read_response = filename.pipe(
        ops.map(lambda i: file.Read(id='config', path=i)),
        file.read(file_response),
    )

    if scheduler is not None:
        read_request = read_request.pipe(
            ops.subscribe_on(scheduler),
        )

    config = read_response.pipe(
        ops.filter(lambda i: i.id == "config"),
        ops.flat_map(lambda i: i.data),
        parse_config(),
    )
    return config, read_request


def read_config_from_consul(kv_adapter, endpoint, key):

    value = kv_adapter.api.watch_key(endpoint, key).pipe(
        ops.map(lambda i: i.value),
        parse_config(),
    )

    return value


def identity_or_redirect(config, kv_adapter):
    '''Returns the configuration data as is, or the content from a redirection.

    Args:
        config: Dict containing the configuration content.
        kv_adapter: Consul adapter, used to monitor the configuration.

    Returns:
        Observable of configuration items
        Observable of http requests
    '''
    if 'redirect' in config and config['redirect']['connector'] == 'consul':
        return read_config_from_consul(
            kv_adapter,
            config['redirect']['endpoint'],
            config['redirect']['key'])
    else:
        return rx.just(config)


def read_config_from_args(args,
                          file_response, http_response,
                          scheduler=None):
    args = parse_arguments(args)
    filename = args.pipe(
        ops.filter(lambda i: i.key == 'config'),
        ops.map(lambda i: i.value),
    )

    config, read_request = read_config_from_file(
        filename, file_response,
        scheduler)

    kv_adapter = kv.adapter(http_response)
    config = config.pipe(
        ops.flat_map(lambda i: identity_or_redirect(i, kv_adapter))
    )

    return config, read_request, kv_adapter.sink



from makinage.serve.serve import create_transform_functions, infer

import numpy as np


def test_create_transform_default():
    c = {'config': {'serve': {}}}
    t = create_transform_functions(c)
    assert t.pre == np.array
    assert t.post == list


def test_predict():
    predict_count = 0

    def _predict(i):
        nonlocal predict_count
        predict_count += 1
        return np.array([i+1])

    c = {'config': {'serve': {
        'pre_transform': 'makinage.sample.serve:pre_transform',
        'post_transform': 'makinage.sample.serve:post_transform',
        'input_field': 'x', 'output_field': 'pred',
    }}}
    t = create_transform_functions(c)
    data = {'x': 42}
    actual_result = infer(data, t, _predict)

    assert predict_count == 1
    assert actual_result == {'x': 42, 'pred': [43]}



import asyncio
from aiokafka import AIOKafkaConsumer, TopicPartition


async def pull(loop, server, topic, group_id, batch_size=1, shuffle=False):
    client = AIOKafkaConsumer(
        topic,
        loop=loop,
        bootstrap_servers=server,
        group_id=group_id,
        auto_offset_reset='earliest',
        enable_auto_commit=False,
    )
    await client.start()

    partitions = client.partitions_for_topic(topic)
    while partitions is None:
        await asyncio.sleep(0.1)

    partitions = list(partitions)
    partitions = [TopicPartition(topic, partition) for partition in partitions]
    #current_offsets = await client.beginning_offsets(partitions)
    end_offsets = await client.end_offsets(partitions)
    current_partition = 0
    done = False

    async def next_partition(current_partition):
        current_partition += 1  # todo recursive
        if current_partition >= len(partitions):
            return None

        current_offset = await client.position(partitions[current_partition])
        if current_offset >= end_offsets[partitions[current_partition]]:
            current_partition = await next_partition(current_partition)
        print("remaining record: {}, partition: {}".format(remaining_records, current_partition))
        return current_partition

    current_offset = await client.position(partitions[current_partition])
    if current_offset >= end_offsets[partitions[current_partition]]:
        done = True

    while done is False:
        remaining_records = batch_size
        batch = []
        while remaining_records > 0:
            msg = await client.getone(partitions[current_partition])
            batch.append(msg)
            remaining_records -= 1

            current_offset = await client.position(partitions[current_partition])
            if current_offset >= end_offsets[partitions[current_partition]]:
                current_partition = await next_partition(current_partition)
                print("remaining record: {}, partition: {}".format(remaining_records, current_partition))
                if current_partition is None:
                    done = True
                    break

        if len(batch) > 0:
            yield(batch)

        '''
        data = await client.getmany(max_records=batch_size)
        print(data)
        #for tp, messages in data.items():
        messages = data[topic]
        if len(messages) > 0:
            batch = []
            for msg in messages:
                batch.append(msg)
            yield(batch)
        else:
            done = True
        '''

    await client.stop()



import json


def encoder():
    ''' Json encoder

    encodes any object using the JSON format
    '''
    def _encode(i):
        return json.dumps(i).encode()

    def _decode(i):
        return json.loads(i.decode())

    return _encode, _decode



from makinage.encoding.none import encoder

def test_none_encoder():
    encode, decode = encoder()

    data = b'azertyuiopqsdfghjll123'
    assert encode(data) == data
    assert decode(data) == data



import asyncio
from collections import namedtuple
from functools import partial

import rx
import rx.operators as ops
from rx.scheduler.eventloop import AsyncIOScheduler

from cyclotron import Component
from cyclotron.debug import trace_observable
from cyclotron.asyncio.runner import run
import cyclotron_std.sys.argv as argv
import cyclotron_std.io.file as file
import cyclotron_aiohttp.http as http

from .config import read_config_from_args
from .operator import create_operators

import cyclotron_aiokafka as kafka

MakiNageSink = namedtuple('MakiNageSink', [
    'kafka', 'http', 'file',
])
MakiNageSource = namedtuple('MakiNageSource', [
    'kafka', 'http', 'file', 'argv',
])
MakiNageDrivers = namedtuple('MakiNageDrivers', [
    'kafka', 'http', 'file', 'argv'
])


Values = namedtuple('Values', ['id', 'observable'])


def makinage(aio_scheduler, sources):
    config, read_request, http_request = read_config_from_args(
        sources.argv.argv,
        sources.file.response,
        sources.http.response,
        scheduler=aio_scheduler
    )

    first_config = rx.concat(config.pipe(ops.take(1),), rx.never())

    kafka_source = sources.kafka.response.pipe(
        trace_observable("kafka source1"),
        ops.replay(),
        ops.ref_count(),
        trace_observable("kafka source2"),
    )
    kafka_source.subscribe()

    kafka_request = first_config.pipe(
        ops.flat_map(lambda i: create_operators(
            i, config,
            kafka_source,
            sources.kafka.feedback.pipe(ops.share()),
        )),
        ops.subscribe_on(aio_scheduler),
        trace_observable("makinage"),
    )

    '''
    config.pipe(ops.subscribe_on(aio_scheduler)).subscribe(
        on_next=print,
        on_error=print,
    )
    '''

    return MakiNageSink(
        file=file.Sink(request=read_request),
        http=http.Sink(request=http_request),
        kafka=kafka.Sink(request=kafka_request),
    )


def main():
    loop = asyncio.get_event_loop()
    # loop.set_debug(True)
    aio_scheduler = AsyncIOScheduler(loop=loop)
    run(
        Component(
            call=partial(makinage, aio_scheduler),
            input=MakiNageSource),
        MakiNageDrivers(
            kafka=kafka.make_driver(),
            http=http.make_driver(),
            file=file.make_driver(),
            argv=argv.make_driver(),
        ),
        loop=loop,
    )



from makinage.operator import initialize_topics


def test_init_topics():
    config_topics = [
        {'name': 'foo'},
        {'name': 'bar', 'encoder': 'makinage.encoding.string'},
    ]
    topics = initialize_topics(config_topics)
    assert b'test' == topics['foo'].encode('test')
    assert 'test' == topics['foo'].decode(b'test')

    assert b'test' == topics['bar'].encode('test')
    assert 'test' == topics['bar'].decode(b'test')



import rx.operators as ops
import rxsci
from cyclotron.debug import trace_observable


def multiply(config, values):
    sink = values.pipe(
        trace_observable("multiply 1"),
        rxsci.with_latest_from(config),
        ops.starmap(lambda i, c: int(i) * c['config']['multiply']),
        trace_observable("multiplied"),
        ops.map(lambda i: str(i))
    )
    return sink,



import sys

try:
    from setuptools import setup, find_packages
    use_setuptools = True
except ImportError:
    from distutils.core import setup
    use_setuptools = False

try:
    with open('README.rst', 'rt') as readme:
        description = '\n' + readme.read()
except IOError:
    # maybe running setup.py from some other dir
    description = ''

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

python_requires = '>=3.6'
install_requires = [
    'Rx>=3.0',
    'PyYAML>=5.3',
    'cyclotron>=1.2',
    'cyclotron-std>=1.0',
    'cyclotron-aiokafka>=0.3',
    'cyclotron-consul>=0.1',
]

setup(
    name="makinage",
    version='0.4.0',
    url='https://github.com/maki-nage/makinage.git',
    license='MIT',
    description="Reactive data science",
    long_description=description,
    author='Romain Picard',
    author_email='romain.picard@oakbits.com',
    packages=find_packages(),
    install_requires=install_requires,
    setup_requires=pytest_runner,
    tests_require=['pytest>=5.0.1'],
    platforms='any',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    project_urls={
        'Documentation': 'https://makinage.readthedocs.io',
    },
    entry_points={
        'console_scripts': [
            'makinage=makinage.makinage:main',
            'makinage-model-publisher=makinage.model_publisher.model_publisher:main',
        ],
    }
)



from importlib import import_module


def import_function(spec):
    spec = spec.split(':')
    module = spec[0]
    fn = spec[1]
    module = import_module(module)
    fn = getattr(module, fn)
    return fn





def pre_transform(config):
    field = config['config']['serve']['input_field']

    def _transform(i):
        return i[field]

    return _transform


def post_transform(config):
    field = config['config']['serve']['output_field']

    def _transform(i, pred):
        i[field] = pred
        return i

    return _transform



import zipfile
import io
import tempfile
import traceback
from collections import namedtuple


import rx.operators as ops
import rxsci as rs
from cyclotron.debug import trace_observable
from makinage.util import import_function

from mlflow.pyfunc import load_model
from mlflow.pyfunc.backend import PyFuncBackend
import numpy as np

Transforms = namedtuple('Transforms', ['pre', 'post'])


def load_mlflow_model(data):
    with tempfile.TemporaryDirectory() as tmp:
        data = io.BytesIO(data)
        with zipfile.ZipFile(data) as artifact:
            artifact.extractall(path=tmp)
            model = load_model(tmp)
            return model


def create_model_predict(model):
    print("create_model_predict: {}".format(type(model)))
    try:
        return model.keras_model.predict  # temporary until mlflow #2830
    except Exception:
        return model.predict


def infer(data, transforms, predict):
    try:
        pre_data = transforms.pre(data)
        prediction = predict(pre_data)
        prediction = transforms.post(data, prediction)
        return prediction
    except Exception as e:
        print("infer error: {}".format(e))
        print(traceback.print_tb(e.__traceback__))
        return None


def create_transform_functions(config):
    if 'pre_transform' in config['config']['serve']:
        pre_transform = import_function(config['config']['serve']['pre_transform'])
        pre_transform = pre_transform(config)
    else:
        pre_transform = np.array

    if 'post_transform' in config['config']['serve']:
        post_transform = import_function(config['config']['serve']['post_transform'])
        post_transform = post_transform(config)
    else:
        post_transform = list

    return Transforms(pre_transform, post_transform)


def serve(config, model, data):
    '''Serves a model

    This operator serves a model. It loads models received on the model
    observable, and executes it on each item received on the data observable. 

    The configuration observable must contain a serve section with the following
    fields:

    * input_field: The input field name used to run inference.
    * output_field: The output field name where inference result is set.

    additionally, a "prepare" field can be set if some data transformation is
    needed before feeding the model. When not present, the input data is
    converted to a numpy array

    Args:
        config: configuration observable.

    Returns:
        An observable of predictions. Each item is a copy of the original datay
        item, with an additional field. The name of the additional field if the
        one set in output_field.
    '''
    predict = model.pipe(
        trace_observable(prefix="model", trace_next_payload=False),
        ops.map(load_mlflow_model),
        ops.map(create_model_predict),
    )

    transforms = config.pipe(
        trace_observable(prefix="prepare", trace_next_payload=False),
        ops.map(create_transform_functions)
    )

    prediction = data.pipe(
        rs.with_latest_from(transforms, predict),
        ops.starmap(infer),
        ops.filter(lambda i: i is not None),
    )

    return prediction,
